#![forbid(unsafe_code)]

use std::io::{stdout, Write, Read, stdin};

pub fn eval_mem(bf_str: &&str, mem: &mut [u8], pointer: &mut usize) -> Result<(), String> {

    let mut pos = 0 as usize;
    let progsize = bf_str.len();
    let chars = bf_str.chars().collect::<Vec<char>>();
    
    let mut loop_stack: Vec<usize> = vec![];

    loop {
        if pos == progsize {
            break
        }

        let char = chars[pos];
        
        pos += 1;

        if char == '>' {
            if (*pointer) + 1 == 30000 {
                return Err("Attempt to increase pointer exceeding memory size".into());
            }
            *pointer += 1;
            continue
        }
        if char == '<' {
            if *pointer == 0 {
                return Err("Attempt to reduce pointer to less than 1".into());
            }
            *pointer -= 1;
            continue
        }
        if char == '+' {
            if mem[*pointer] == 255 {
                mem[*pointer] = 0;
                continue
            }
            mem[*pointer] += 1;
            continue
        }
        if char == '-' {
            if mem[*pointer] == 0 {
                mem[*pointer] = 255;
                continue
            }
            mem[*pointer] -= 1;
            continue
        }
        if char == '.' {
            print!("{}", mem[*pointer] as char);
            stdout().flush().unwrap();
            continue
        }
        if char == ',' {
            let byte = stdin().bytes().next();
            if byte.is_none() {
                return Err("Stdin closed".into());
            }
            let byte = byte.unwrap();
            if byte.is_err() {
                return Err(byte.unwrap_err().to_string());
            }
            mem[*pointer] = byte.unwrap();
            continue
        }
        if char == '[' {
            loop_stack.push(pos);
            continue
        }
        if char == ']' {

            if loop_stack.len() == 0 {
                return Err(format!("Invalid loop at position {pos}"));
            }

            if mem[*pointer] == 0 {
                loop_stack.pop();
                continue
            } else {
                pos = loop_stack.last().unwrap().clone();
                continue
            }
        }
    }

    if loop_stack.len() != 0 {
        return Err(format!("Unclosed loop at position {}", loop_stack.last().unwrap()));
    }

    Ok(())
}

pub fn eval(bf_str: &&str) -> Result<[u8; 30000], String> {
    let mut memory: [u8; 30000] = core::array::from_fn(|_| 0);
    let mut pointer = 0 as usize;

    let e = eval_mem(bf_str, &mut memory, &mut pointer);
    if e.is_err() {
        return Err(e.unwrap_err());
    }

    Ok(memory)
}